#!/usr/bin/env python3
# coding=utf-8
from itertools import cycle
from statistics import mean
from typing import Iterable

from .core import *
from ..simulator import Sim, SchedulingFailure


class SFF(object):
    @Sim.register_reset_global_fields
    class Props:
        def __init__(self):
            self.counter_packets_put_on_wire: int = 0
            self.consider_link_capacity: bool = None
            self.linkBwCap: List[List[int]] = None
            self.linkBwRemaining: List[List[int]] = None
            self.linkLatency: List[List[int]] = None
            self.latencyProvider: Dict[int, Iterable[int]] = None
            
            self.end_to_end_latency: List[List[int]] = None
            self.end_to_end_bw: List[List[int]] = None
            self.end_to_end_next_hop: List[List[int]] = None
            
            self.allSFFs: Dict[int, 'SFF'] = dict()
    
    @staticmethod
    @Sim.register_simulation_start_hook
    def init_data_structure(sim: Sim):
        # make sure that we reset the data structure only if we are already at the beginning of the simulation,
        # of if somebody calls this function explicitly (before starting the simulation)
        sff_props = sim.props.sff
        if not sim.run or sff_props.linkLatency is None:
            di = len(sff_props.allSFFs)
            sff_props.linkLatency = [[0 for _ in range(di)] for _ in range(di)]
            sff_props.linkBwRemaining = [[0 for _ in range(di)] for _ in range(di)]
            sff_props.linkBwCap = [[0 for _ in range(di)] for _ in range(di)]
            
            sff_props.end_to_end_latency = [[0 for _ in range(di)] for _ in range(di)]
            sff_props.end_to_end_bw = [[0 for _ in range(di)] for _ in range(di)]
            sff_props.end_to_end_next_hop = [[0 for _ in range(di)] for _ in range(di)]
    
    @staticmethod
    @Sim.register_stop_sim_catch_all_packets_hooks
    def callback_get_all_packets_from_sff(sim: Sim):
        for sff in sim.props.sff.allSFFs.values():
            if sff.scheduler.requires_queues_per_class():
                for q in sff.packet_queue_per_class:
                    for p in sff.packet_queue_per_class[q]:
                        p.timeQueueScheduling += p.get_delta_of_time_mark()
                        yield p
            else:
                for p in sff.packet_queue:
                    p.timeQueueScheduling += p.get_delta_of_time_mark()
                    yield p
    
    def __repr__(self):
        return f"SFF({str(self.id)}/{self.get_number_of_queued_packets()})"
    
    def __init__(self, sim: Sim, scheduler):
        if sim.props.sff.linkLatency is not None:
            raise NameError(
                "Cannot create new SFF after initializing the data structures. " +
                "You have to create all SFFs, and then setup the connections.")
        
        self.id = len(sim.props.sff.allSFFs)
        sim.props.sff.allSFFs[self.id] = self
        self.sim = sim
        
        self.scheduler = scheduler
        self.SFIsPerType: Dict[int, List['SFI']] = dict()
        self.servers = set()
        
        self.service_rate_per_sf: Dict[int, float] = {}
        
        # this is either a single queue, or a dictionary of queues
        if self.scheduler.requires_queues_per_class():
            self.packet_queue_per_class: Dict[int, deque] = dict()
        else:
            self.packet_queue: deque = deque()
        
        scheduler.assign_sff(self)
        
        # holds all queued events when the outgoing link does not provide
        # enough capacity
        self.outQueue = dict()
    
    @staticmethod
    def init_end_to_end_paths(sim: Sim):
        di = len(sim.props.sff.allSFFs)
        # initialize with direct edges
        for s in range(di):
            for d in range(di):
                assert (sim.props.sff.end_to_end_bw[s][d] == 0)
        
        # draw randomly 500 latencies and calculate the mean
        expected_latencies = [mean([next(sim.props.sff.latencyProvider[dist]) for _ in range(500)])
                              for dist in sim.props.sff.latencyProvider]
        
        for s in range(di):
            for d in range(di):
                if sim.props.sff.linkBwCap[s][d] > 0:
                    sim.props.sff.end_to_end_bw[s][d] = sim.props.sff.linkBwCap[s][d]
                    sim.props.sff.end_to_end_latency[s][d] = expected_latencies[sim.props.sff.linkLatency[s][d]]
                    sim.props.sff.end_to_end_next_hop[s][d] = d
        
        # this is simply floyd warshall
        for via in range(di):
            for s in range(di):
                for d in range(di):
                    if s == d:
                        continue
                    if s == via or d == via:
                        continue
                    # check if s->d using "via" is a valid connection
                    if sim.props.sff.end_to_end_bw[s][via] > 0 and sim.props.sff.end_to_end_bw[via][d] > 0:
                        # faster?
                        new_latency = sim.props.sff.end_to_end_latency[s][via] + sim.props.sff.end_to_end_latency[via][
                            d]
                        if sim.props.sff.end_to_end_bw[s][d] == 0 or new_latency < sim.props.sff.end_to_end_latency[s][
                            d]:
                            # either new latency is faster, or there was no connection before
                            sim.props.sff.end_to_end_latency[s][d] = new_latency
                            sim.props.sff.end_to_end_next_hop[s][d] = sim.props.sff.end_to_end_next_hop[s][via]
                            sim.props.sff.end_to_end_bw[s][d] = min(sim.props.sff.end_to_end_bw[s][via],
                                                                    sim.props.sff.end_to_end_bw[via][d])
    
    def get_number_of_queued_packets(self):
        if self.scheduler.requires_queues_per_class():
            return sum(map(len, self.packet_queue_per_class.values()))
        else:
            return len(self.packet_queue)
    
    def route_packet_to_sfi(self, packet: 'Packet', sfi: 'SFI'):
        if self.sim.DEBUG:
            print("route packet to sfi " + str(sfi))
        
        packet.mark_time()
        sfi_props: 'SFI.Props' = self.sim.props.sfi
        sff_props: SFF.Props = self.sim.props.sff
        
        delay = next(sff_props.latencyProvider[sfi_props.latency_provider])
        self.sim.schedule_event(NetworkDelayEvent(delay=delay,
                                                  inner_packet=packet,
                                                  source=self,
                                                  dest_id=sfi.id,
                                                  source_is_sff=True,
                                                  dest_is_sff=False))
    
    def free_bw_resource_to_dest_id(self, dest_id, packet: Packet):
        assert self.sim.props.sff.consider_link_capacity
        self.sim.props.sff.linkBwRemaining[self.id][dest_id] += packet.transmission_size
    
    # sends a packet to a SFF
    def route_packet_to_sff_id(self, packet: Packet, dest_id):
        if self.sim.DEBUG:
            print("route packet to sff" + str(dest_id))
        # mark current time for statistics
        packet.mark_time()
        
        sff_props: SFF.Props = self.sim.props.sff
        
        if sff_props.linkBwCap[self.id][dest_id] < packet.transmission_size:
            print("end 2 end links says capacity:{0} using next hop {1}".
                  format(sff_props.end_to_end_bw[self.id][dest_id], sff_props.end_to_end_next_hop[self.id][dest_id]))
            raise NameError(f"cannot route this packet {self.id}->{dest_id}, because bw capacity " +
                            f"({sff_props.linkBwCap[self.id][dest_id]}) is not sufficient for packet " +
                            f"of size ({packet.transmission_size})!")
        if sff_props.consider_link_capacity:
            if not (dest_id in self.outQueue):
                self.outQueue[dest_id] = []
            if (len(self.outQueue[dest_id]) == 0 and packet.transmission_size <=
                    sff_props.linkBwRemaining[self.id][dest_id]):
                self.put_packet_on_wire(packet=packet, dest_id=dest_id)
            else:
                # enqueue packet
                self.outQueue[dest_id].append(packet)
        else:
            self.put_packet_on_wire(packet=packet, dest_id=dest_id)
    
    # internal method, don't call directly
    # this method does not check if some packet are in the out queue
    # use route_packet_to_sff_id() for sending packets to a SFF
    def put_packet_on_wire(self, packet: Packet, dest_id):
        packet.timeQueueNetwork += packet.get_delta_of_time_mark()
        packet.mark_time()
        sff_props: SFF.Props = self.sim.props.sff
        # enough bw so that we can send the packet immediately
        delay = next(sff_props.latencyProvider[sff_props.linkLatency[self.id][dest_id]])
        self.sim.schedule_event(NetworkDelayEvent(delay=delay,
                                                  inner_packet=packet, source=self, dest_id=dest_id))
        if sff_props.consider_link_capacity:
            sff_props.linkBwRemaining[self.id][dest_id] -= packet.transmission_size
    
    def route_packet_to_next_hop(self, packet: Packet):
        next_hop_type, next_hop = packet.fullPath[packet.pathPosition]
        packet.pathPosition += 1
        
        if self.sim.PACKET_ID_TO_DEBUG == packet.id:
            print(f'** packet at sff {self.id}, route to next hop')
        
        if next_hop_type == 'SFF':
            if self.sim.DEBUG:
                print(". packet goes to another SFF")
            # we send this packet to the other SFF
            if packet.id == self.sim.PACKET_ID_TO_DEBUG:
                print("** debug packet route to {0}".format(str(next_hop)))
            self.route_packet_to_sff_id(packet, next_hop.id)
        elif next_hop_type == 'SFI':
            if self.sim.DEBUG:
                print(". packet goes to a SFI")
            # we send this packet to the SFI
            self.route_packet_to_sfi(packet, next_hop)
        else:
            NameError(f"unknown next hop: {next_hop_type}")
    
    def check_and_update_packet_and_return_if_process_locally(self, packet: Packet, source: str) -> bool:
        if self.sim.TRACE_PACKET_PATH:
            packet.visitedHops.append(self)
        if self.sim.DEBUG:
            print(f'at SFF {self.id}, receive a packet from {source}; '
                  f'packet: id:{packet.id} sfc:{packet.flow.sfTypeChain} '
                  f'remaining path:{packet.fullPath[packet.pathPosition:]}')
        if packet.id == self.sim.PACKET_ID_TO_DEBUG:
            print(
                f"** @{self.sim.currentTime} debug packet at {str(self)}, path={packet.fullPath}, coming from {source}")
        if packet.flow.qosMaxDelay < (self.sim.currentTime - packet.time_ingress):
            if self.sim.DEBUG:
                print(". drop packet because of timeout")
            # print(". drop packet because of timeout " + source)
            
            packet.drop_timed_out(self)
            return False
        
        if packet.processing_done:
            # do we have to add the path to the egress?
            if len(packet.toBeVisited) == 0 and len(
                    packet.fullPath) == packet.pathPosition and self.id != packet.flow.desiredEgressSSFid:
                # add the path to the egress
                path_to_dest = self.get_multi_hop_path_for(self.sim, self.id, packet.flow.desiredEgressSSFid)
                for sff_id in path_to_dest:
                    packet.fullPath.append((SFF.__name__, self.sim.props.sff.allSFFs[sff_id]))
            
            # we reached already the egress?
            if len(packet.fullPath) == packet.pathPosition:
                if self.sim.DEBUG:
                    print(". packet reached egress")
                
                # we should be at the egress
                assert (self.id == packet.flow.desiredEgressSSFid)
                # there should be no remaining sf type
                assert (len(packet.toBeVisited) == 0)
                
                packet.done()
                return False
            else:
                self.route_packet_to_next_hop(packet)
                return False
        elif len(packet.fullPath) > packet.pathPosition:
            self.route_packet_to_next_hop(packet)
            return False
        
        return True
    
    def handle_packet_from_ingress(self, packet: Packet):
        if self.check_and_update_packet_and_return_if_process_locally(packet, f'ingress'):
            self.apply_logic_packet_from_ingress(packet)
    
    def handle_packet_from_other_sff(self, packet: Packet, other_sff: 'SFF'):
        if self.check_and_update_packet_and_return_if_process_locally(packet, f'a sff {other_sff.id}'):
            self.apply_logic_packet_from_other_sff(packet, other_sff)
    
    def handle_packet_from_sfi(self, packet: Packet, sfi: 'SFI'):
        if self.check_and_update_packet_and_return_if_process_locally(packet, f'a sfi {sfi.id}'):
            self.apply_logic_packet_from_sfi(packet, sfi)
    
    def handle_packet_from_scheduler(self, packet: Packet):
        if self.sim.PACKET_ID_TO_DEBUG == packet.id:
            print(f'** receive packet from scheduler..')
            Packet.debug_print_path(packet.fullPath, packet.pathPosition)
        
        if self.check_and_update_packet_and_return_if_process_locally(packet, 'scheduler'):
            Packet.debug_packet(packet)
            raise NameError(f'scheduler returns a packet {packet.id}, but sff does not know how to proceed with this '
                            f'packet')
    
    def apply_logic_packet_from_other_sff(self, packet, other_sff: 'SFF'):
        self.put_packet_in_queue(packet)
        self.inform_scheduler_about_packet(packet)
    
    def apply_logic_packet_from_ingress(self, packet):
        self.put_packet_in_queue(packet)
        self.inform_scheduler_about_packet(packet)
    
    def apply_logic_packet_from_sfi(self, packet, sfi: 'SFI'):
        self.put_packet_in_queue(packet)
        self.inform_scheduler_about_packet(packet)
    
    def inform_scheduler_about_packet(self, packet):
        if self.sim.DEBUG or packet.id == self.sim.PACKET_ID_TO_DEBUG:
            print(f'** send packet ({packet.id}) to scheduler')
        packet.seenByScheduler += 1
        try:
            self.scheduler.handle_packet_arrival(packet=packet)
        except SchedulingFailure as e:
            if self.sim.STOP_SIMULATION_IF_SCHEDULER_WAS_UNSUCCESSFUL:
                raise e
            elif self.sim.DEBUG:
                print(
                    ". scheduler was unsuccessful to schedule a packet: " +
                    str(e))
    
    def put_packet_in_queue(self, packet):
        if self.scheduler.requires_queues_per_class():
            queue = Flow.get_packet_class_of_packet(packet)
            
            if queue not in self.packet_queue_per_class:
                self.packet_queue_per_class[queue] = deque()
            if self.sim.DEBUG:
                print(f'. add packet of sfc {packet.flow.sfc_identifier} '
                      f'at position {packet.sfc_position} to queue {queue}')
            self.packet_queue_per_class[queue].append(packet)
        else:
            self.packet_queue.append(packet)
        
        packet.mark_time()  # mark time when this packet was queued to the scheduler
    
    def register_sfi(self, sfi: 'SFI'):
        if not (sfi.of_type in self.SFIsPerType):
            self.SFIsPerType[sfi.of_type] = []
        # check here that we don't get for a server multiplel SFI of the same
        # type
        for otherSFI in self.SFIsPerType[sfi.of_type]:
            if otherSFI.server == sfi.server:
                raise NameError(
                    "this configuration is not allowed! A SFF should" +
                    " see for a single server only one SFI instance per type!")
        self.SFIsPerType[sfi.of_type].append(sfi)
        self.servers.add(sfi.server)
        
        if sfi.of_type not in self.service_rate_per_sf:
            self.service_rate_per_sf[sfi.of_type] = 0
        self.service_rate_per_sf[sfi.of_type] += sfi.get_expected_processing_rate()
    
    def sfi_finishes_processing_of_packet(self, sfi: 'SFI', packet: Packet):
        # the given SFI is finished with processing, so depeding on the server cpu sharing policy,
        # this server might be free now
        self.scheduler.notify_sfi_finished_processing_of_packet(sfi, packet)
    
    @staticmethod
    def get_multi_hop_bw_for(sim: Sim, source_id, dest_id):
        sff_props: SFF.Props = sim.props.sff
        if sff_props.end_to_end_bw[0][1] == 0:
            SFF.init_end_to_end_paths(sim)
        
        if sff_props.end_to_end_bw[source_id][dest_id] == 0:
            raise NameError("graph is not connected")
        
        return sff_props.end_to_end_bw[source_id][dest_id]
    
    @staticmethod
    def get_multi_hop_latency_for(sim: Sim, source_id, dest_id):
        if source_id == dest_id:
            return 0
        
        sff_props: SFF.Props = sim.props.sff
        if sff_props.end_to_end_bw[0][1] == 0:
            SFF.init_end_to_end_paths(sim)
        
        if sff_props.end_to_end_bw[source_id][dest_id] == 0:
            raise NameError("graph is not connected")
        
        return sff_props.end_to_end_latency[source_id][dest_id]
    
    @staticmethod
    def get_next_hop_for(sim: Sim, source_id, dest_id):
        sff_props: SFF.Props = sim.props.sff
        if sff_props.end_to_end_bw[0][1] == 0:
            SFF.init_end_to_end_paths(sim)
        
        if sff_props.end_to_end_bw[source_id][dest_id] == 0:
            raise NameError("graph is not connected")
        
        return sff_props.end_to_end_next_hop[source_id][dest_id]
    
    @staticmethod
    def get_multi_hop_route_between_ids(sim: Sim, from_id, to_id):
        if from_id == to_id:
            return []
        steps = SFF.get_multi_hop_path_for(sim, from_id, to_id)
        all_sff: SFF.Props = sim.props.sff.allSFFs
        return [(SFF.__name__, all_sff[step]) for step in steps]
    
    @staticmethod
    def get_multi_hop_path_for(sim: Sim, source_id, dest_id):
        sff_props: SFF.Props = sim.props.sff
        if sff_props.end_to_end_bw[0][1] == 0:
            SFF.init_end_to_end_paths(sim)
        
        if sff_props.end_to_end_bw[source_id][dest_id] == 0:
            raise NameError(f"graph is not connected; failed to find path for {source_id}->{dest_id}")
        
        path = []
        
        next_hop = sff_props.end_to_end_next_hop[source_id][dest_id]
        path.append(next_hop)
        
        while next_hop != dest_id:
            next_hop = sff_props.end_to_end_next_hop[next_hop][dest_id]
            path.append(next_hop)
        
        return path
    
    @staticmethod
    def check_if_connection_exists(sim: Sim, source_id, dest_id):
        if sim.props.sff.linkBwCap[source_id][dest_id] > 0:
            return True
        return False
    
    @staticmethod
    def setup_latency_distribution(sim: Sim, id: int, values: List[int]):
        sff_props: SFF.Props = sim.props.sff
        if sff_props.latencyProvider is None:
            sff_props.latencyProvider = {}
        sff_props.latencyProvider[id] = cycle(values)
    
    @staticmethod
    def setup_connection(sim: Sim, source_id, destination_id,
                         bw_cap, latency_provider: int, bidirectional: bool = False):
        sff_props: SFF.Props = sim.props.sff
        # do we have to setup the data structures?
        if sff_props.linkLatency is None:
            SFF.init_data_structure(sim)
        assert (len(sff_props.linkLatency) > source_id)
        assert (len(sff_props.linkLatency) > destination_id)
        assert (source_id != destination_id)
        assert sff_props.latencyProvider is not None
        if latency_provider not in sff_props.latencyProvider:
            raise NameError(f"unknown latency provider id given: {latency_provider}")
        
        # bw_cap == 0, means that there is no connection!
        sff_props.linkBwCap[source_id][destination_id] = int(bw_cap)
        sff_props.linkLatency[source_id][destination_id] = latency_provider
        sff_props.linkBwRemaining[source_id][destination_id] = int(bw_cap)
        if bidirectional:
            sff_props.linkBwCap[destination_id][source_id] = int(bw_cap)
            sff_props.linkLatency[destination_id][source_id] = latency_provider
            sff_props.linkBwRemaining[destination_id][source_id] = int(bw_cap)
    
    @staticmethod
    def get_delay_of_connection(src: 'SFF', dest: 'SFF') -> int:
        if src == dest:
            return 0
        assert src.sim == dest.sim
        return SFF.get_multi_hop_latency_for(src.sim, src.id, dest.id)
    
    @staticmethod
    def set_consider_bw_capacity(sim: Sim, consider_bandwidth_capacity_of_links: bool):
        sim.props.sff.consider_link_capacity = consider_bandwidth_capacity_of_links



#!/usr/bin/env python3
# coding=utf-8
from collections import deque
from enum import unique, Enum
from typing import Dict, List, Tuple

from ..events import PacketHoldingEvent
from ..simulator import Sim


@unique
class ServerCpuPolicy(Enum):
    static = 1
    dynamic = 2
    one_at_a_time = 3


class Flow(object):
    @Sim.register_reset_global_fields
    class Props:
        def __init__(self):
            # parameters: the sf type chain configuration this flow belongs to
            self.lastId: int = 0
            
            # the sfc id is simply the actual sfc written as a string
            # the sfc class is an integer which refers to the buffer/class of this sfc of position 0.
            #   so when this sfc has 3 entries, it reserves buffer/class "sfc class", "sfc class"+1, "sfc class"+2
            
            # holds a map from sfc id to sfc class
            self.sfc_classes: Dict[str, int] = dict()
            
            self.statistics_drops_per_class: Dict[int, int] = dict()
            
            # holds a map from sfc class to deadline of this sfc
            self.sfc_class_to_deadline: Dict[int, int] = dict()
            self.max_deadline: int = None
            
            # holds a map from sfc class to egress id of this sfc
            self.sfc_class_to_egress: Dict[int, int] = dict()
            
            # holds a map from sfc class to actual (sf, bool: last_sf_in_sfc)
            self.sfc_class_to_sf: Dict[int, Tuple[int, bool]] = dict()
            
            # holds the next class which is not used so far
            self.sfc_next_free_class = 0
            
            # if set true, the same SFC but with different egress, will end up in different classes
            self.individual_class_per_egress: bool = False
    
    @staticmethod
    def get_sfc_identifier(sim: Sim, sfc_type: list, egress_id: int):
        if sim.props.flow.individual_class_per_egress:
            return "-".join(str(c) for c in sfc_type) + f":{egress_id}"
        else:
            return "-".join(str(c) for c in sfc_type)
    
    @staticmethod
    def register_sfc_for_packet_classes(sim: Sim, sfc_type: list, qos_max_delay: int, egress_id: int):
        sfc_identifier = Flow.get_sfc_identifier(sim, sfc_type, egress_id)
        assert sfc_identifier not in sim.props.flow.sfc_classes
        
        sim.props.flow.sfc_classes[sfc_identifier] = sim.props.flow.sfc_next_free_class
        sim.props.flow.sfc_next_free_class += len(sfc_type)
        
        if sim.run:
            raise NameError("it is not allowed to create new flows of NEW SFC/egress combinations after " +
                            "starting the simulation! (because of caching)")
        
        if sim.DEBUG:
            print(f"..register sfc identifier {sfc_identifier} for packet class "
                  f"{sim.props.flow.sfc_classes[sfc_identifier]}")
        
        for i in range(len(sfc_type)):
            this_class_identifier = sim.props.flow.sfc_classes[sfc_identifier] + i
            sim.props.flow.sfc_class_to_sf[this_class_identifier] = (
                sfc_type[i], (i + 1) == len(sfc_type))
            if this_class_identifier in sim.props.flow.sfc_class_to_deadline:
                raise NameError("we store the qos delay per class, which is used by the optimal scheduler, however," +
                                f" your setup has different values for the same SFC {sfc_identifier}")
            sim.props.flow.sfc_class_to_deadline[this_class_identifier] = qos_max_delay
            if sim.props.flow.individual_class_per_egress:
                sim.props.flow.sfc_class_to_egress[this_class_identifier] = egress_id
            
            for sff in sim.props.sff.allSFFs.values():
                if sff.scheduler.requires_queues_per_class():
                    sff.packet_queue_per_class[this_class_identifier] = deque()
            sim.props.flow.statistics_drops_per_class[this_class_identifier] = 0
    
    @staticmethod
    def get_packet_class_of_packet(packet: 'Packet'):
        return packet.flow.sim.props.flow.sfc_classes[packet.flow.sfc_identifier] + packet.sfc_position
    
    @staticmethod
    def get_sf_and_eoc_of_packet(packet: 'Packet') -> tuple:
        return Flow.get_sf_and_eoc_of(packet.flow.sim, packet.flow.sfc_identifier, packet.sfc_position)
    
    @staticmethod
    def get_sf_and_eoc_of(sim: Sim, sfc_identifier: str, at_position: int) -> tuple:
        flow_pros: Flow.Props = sim.props.flow
        return flow_pros.sfc_class_to_sf[flow_pros.sfc_classes[sfc_identifier] + at_position]
    
    @staticmethod
    def debug_get_sfc_identifier_and_pos_of_packet_class(sim: Sim, packet_class: int):
        if not sim.DEBUG:
            # print("WARNING this method is not intended to be used for normal runs")
            pass
        # find the sfc_identifier that has the largest packet_class but still smaller/equal than the given
        flow_pros: Flow.Props = sim.props.flow
        found_sfc = None
        found_packet_class = -1
        for other_sfc_identifier in flow_pros.sfc_classes:
            if found_sfc is None or (
                    packet_class >= flow_pros.sfc_classes[other_sfc_identifier] > found_packet_class):
                found_sfc = other_sfc_identifier
                found_packet_class = flow_pros.sfc_classes[other_sfc_identifier]
        assert found_sfc is not None
        return found_sfc, (packet_class - found_packet_class)
    
    def __init__(self, sim: Sim, sf_type_chain: List[int], qos_max_delay: int,
                 desired_egress_ssf_id: int, ingress_sff_id: int, start_time: int = 0):
        self.sim = sim
        flow_props: Flow.Props = sim.props.flow
        flow_props.lastId += 1
        self.id = flow_props.lastId
        self.sfTypeChain = [int(i) for i in sf_type_chain]
        self.desiredEgressSSFid = desired_egress_ssf_id
        self.ingress_sff_id = int(ingress_sff_id)
        self.qosMaxDelay = qos_max_delay
        self.start_time = start_time
        
        assert len(self.sfTypeChain) > 0
        
        if flow_props.max_deadline is None or flow_props.max_deadline < self.qosMaxDelay:
            flow_props.max_deadline = self.qosMaxDelay
        
        self.sfc_identifier = Flow.get_sfc_identifier(sim, sf_type_chain, desired_egress_ssf_id)
        if self.sfc_identifier not in sim.props.flow.sfc_classes:
            Flow.register_sfc_for_packet_classes(sim, sf_type_chain, qos_max_delay, desired_egress_ssf_id)
        
        self.sfc_class = sim.props.flow.sfc_classes[self.sfc_identifier]
    
    def __str__(self):
        return "Flow(id={0},chain={1},egress={2},qos={3})".format(
            self.id, self.sfTypeChain, self.desiredEgressSSFid, self.qosMaxDelay)


class IngressEvent(PacketHoldingEvent):
    
    def __init__(self, inner_packet: 'Packet', sff_id):
        super().__init__(inner_packet.time_ingress, inner_packet)
        self.sff_id = sff_id
    
    def update_packet_time_tracking(self):
        pass
    
    # send packet to ingress SFF
    def process_event(self):
        sim = self.inner_packet.flow.sim
        sim.last_packet_ingress_time = self.time
        
        stat_props = sim.props.sim_stats
        if stat_props.workloadStats is not None:
            stat_props.workloadStats.add_entry(self.time,
                                               self.inner_packet.flow.id,
                                               '-'.join(map(str, self.inner_packet.flow.sfTypeChain)),
                                               self.sff_id,
                                               self.inner_packet.flow.desiredEgressSSFid,
                                               self.inner_packet.flow.qosMaxDelay)
        
        packet_props = self.inner_packet.flow.sim.props.packet
        packet_props.counter_packet_in_system += 1
        packet_props.counter_ingress_packets += 1
        sim.props.sff.allSFFs[self.sff_id].handle_packet_from_ingress(self.inner_packet)


class Packet(object):
    teardown_hooks = []
    
    @staticmethod
    def register_tear_down(func):
        Packet.teardown_hooks.append(func)
        return func
    
    @Sim.register_reset_global_fields
    class Props:
        def __init__(self):
            self.counter_ingress_packets: int = 0
            self.counter_packet_in_system: int = 0
            
            # We count the number of packets which are still in the simulation after ending the workload, so that
            # we are able to calculate the effective avg delay correctly
            self.counter_packet_after_workload_end_in_system_no_timeout: int = 0
            
            self.all: list = []
            self.statsRatiosQos: float = 0
            self.statsPacketsSuccessfulProcessed: int = 0
            self.statsPacketsRejectedSchedule: int = 0
            self.statsPacketsRejectedProcessingDelay: int = 0
            self.statsSumDelay: int = 0
            self.lastId: int = 0
            self.statsPacketsTotalCount: int = 0
            
            self.dones: list = []
    
    def __init__(self, time_ingress: int, flow: Flow, transmission_size: int):
        self.scheduler_flag = dict()
        packet_props: Packet.Props = flow.sim.props.packet
        if flow.sim.KEEP_LIST_OF_ALL_PACKETS:
            packet_props.all.append(self)
        packet_props.lastId += 1
        self.id = packet_props.lastId
        self.time_ingress = int(time_ingress)
        self.flow = flow
        
        assert flow is not None
        assert self.time_ingress >= self.flow.start_time
        
        self.processing_done = False
        
        # holds the position with respect to the sfc, so position 1 means,
        # that the next SFI to be processed is of type of the second element in the chain
        self.sfc_position = 0
        
        self.pathPosition = 0  # holds the position of the total path
        # holds the full path, some schedulers will build this list
        # incrementally
        self.fullPath = []
        # holds all remaining SF types, that need to be visited
        self.toBeVisited = list(flow.sfTypeChain)
        
        self.ingress_sff_id = flow.ingress_sff_id
        
        self.transmission_size = int(transmission_size)
        self.seenByScheduler = 0
        packet_props.statsPacketsTotalCount += 1
        
        # statistic values
        self.timeQueueProcessing = 0
        self.timeQueueNetwork = 0
        self.timeQueueScheduling = 0
        
        self.timeProcessing = 0
        self.timeNetwork = 0
        self.realTimeScheduling = 0
        
        self.timeMarker = None
        
        self.delay = None
        self.final_state = None
        
        self.callback_when_be_dropped = None
        
        if flow.sim.TRACE_PACKET_PATH:
            self.visitedHops = []
    
    @staticmethod
    @Sim.register_sim_oneliner_text_provider
    def sim_oneline_info(sim: Sim):
        packet_props: Packet.Props = sim.props.packet
        
        avg_succ = (0 if (packet_props.statsPacketsSuccessfulProcessed
                          + packet_props.statsPacketsRejectedProcessingDelay) == 0 else
                    round(packet_props.statsPacketsSuccessfulProcessed
                          / (packet_props.statsPacketsSuccessfulProcessed
                             + packet_props.statsPacketsRejectedProcessingDelay), ndigits=4))
        
        return f'Packets delivered {packet_props.statsPacketsSuccessfulProcessed}/' \
               f'{packet_props.statsPacketsTotalCount} ' \
               f'ingress: {packet_props.counter_ingress_packets} ' \
               f'(timeout:{packet_props.statsPacketsRejectedProcessingDelay} ' \
               f'reject:{packet_props.statsPacketsRejectedSchedule} ' \
               f'avg succ:{avg_succ} ) '
    
    @staticmethod
    def create_wrap_in_event(time_ingress: int, flow: Flow, transmission_size: int) -> 'IngressEvent':
        p = Packet(time_ingress, flow, transmission_size)
        return IngressEvent(inner_packet=p, sff_id=flow.ingress_sff_id)
    
    @staticmethod
    def create_and_schedule(time_ingress: int, flow: Flow, transmission_size: int, ingress_sff_id: int):
        return flow.sim.schedule_event(
            Packet.create_wrap_in_event(time_ingress, flow, transmission_size, ingress_sff_id))
    
    def set_callback_when_dropped(self, callback):
        self.callback_when_be_dropped = callback
    
    # mark the current time, used for statistics purpose
    def mark_time(self):
        assert (self.timeMarker is None)
        self.timeMarker = self.flow.sim.currentTime
    
    # returns the delta of current time and time marker
    def get_delta_of_time_mark(self):
        assert (self.timeMarker is not None)
        tmp = self.flow.sim.currentTime - self.timeMarker
        self.timeMarker = None
        return tmp
    
    def tear_down(self, final_state):
        sim: Sim = self.flow.sim
        packet_props: Packet.Props = sim.props.packet
        
        if sim.PACKET_ID_TO_DEBUG == self.id:
            print(f"Debug packet gets tear down")
            Packet.debug_packet(self)
        
        if sim.KEEP_LIST_OF_ALL_PACKETS:
            packet_props.all.remove(self)
        self.delay = sim.currentTime - self.time_ingress
        self.final_state = final_state
        packet_props.counter_packet_in_system -= 1
        packet_props.statsSumDelay += self.delay
        
        for f in Packet.teardown_hooks:
            f(self)
        
        self.flow = None
        self.callback_when_be_dropped = None
        self.fullPath = None
        self.scheduler_flag = None
    
    def handle_stop_simulation(self):
        # this method will be called when the simulation done, but the packet is somewhere in the simulation
        packet_props: Packet.Props = self.flow.sim.props.packet
        if (self.flow.sim.currentTime - self.time_ingress) >= self.flow.qosMaxDelay:
            self.drop_timed_out(end_of_sim=True)
        else:
            packet_props.counter_packet_in_system -= 1
            packet_props.statsPacketsSuccessfulProcessed += 1
            packet_props.counter_packet_after_workload_end_in_system_no_timeout += 1
    
    def drop_timed_out(self, caller=None, end_of_sim=False):
        if self.id == self.flow.sim.PACKET_ID_TO_DEBUG:
            print(f"** drop debug packet because of timeout, caller:{caller}")
        
        props = self.flow.sim.props
        props.packet.statsPacketsRejectedProcessingDelay += 1
        
        if self.processing_done:
            # this is a very rare case, but the timeout happens during forwarding the packet to the desired egress sff
            # so this means, we would get the wrong packet class, because the pointer in already on the next entry
            class_of_packet = Flow.get_packet_class_of_packet(self) - 1
            props.flow.statistics_drops_per_class[class_of_packet] += 1
        else:
            props.flow.statistics_drops_per_class[Flow.get_packet_class_of_packet(self)] += 1
        
        if (self.callback_when_be_dropped is not None) and (not end_of_sim):
            self.callback_when_be_dropped(self, caller)
            self.callback_when_be_dropped = None
        self.tear_down("timeout")
    
    def reject(self):
        if self.id == self.flow.sim.PACKET_ID_TO_DEBUG:
            print("** reject debug packet")
        
        self.flow.sim.props.packet.statsPacketsRejectedSchedule += 1
        self.tear_down("rejectSchedule")
    
    def done(self):
        packet_props: Packet.Props = self.flow.sim.props.packet
        if self.id == self.flow.sim.PACKET_ID_TO_DEBUG:
            print("** debug packet is done")
        packet_props.statsPacketsSuccessfulProcessed += 1
        
        self.delay = self.flow.sim.currentTime - self.time_ingress
        packet_props.statsRatiosQos += float(self.delay) / self.flow.qosMaxDelay
        
        self.tear_down("done")
    
    @staticmethod
    def debug_packet(packet):
        print("config:")
        print("\t id: {0}".format(str(packet.id)))
        print("\t time ingress: {0}".format(str(packet.time_ingress)))
        print("\t flow: {0}".format(str(packet.flow)))
        print("\t sfc position processed of SFIs: {0}".format(str(packet.sfc_position)))
        print("\t SFs not yet scheduled: {0}".format(str(packet.toBeVisited)))
        print("\t path position: {0}".format(str(packet.pathPosition)))
        print("\t transmission_size: {0}".format(str(packet.transmission_size)))
        print("\t seenByScheduler: {0}".format(str(packet.seenByScheduler)))
        if packet.flow.sim.TRACE_PACKET_PATH:
            print("\t visited hops: {0}".format(
                str(list(map(str, packet.visitedHops)))))
        
        Packet.debug_print_path(packet.fullPath, current_position=packet.pathPosition)
    
    @staticmethod
    def debug_print_path(path: list, current_position=None):
        print("this path has {0} steps".format(len(path)))
        for i, (t, instance) in enumerate(path):
            if current_position is not None and i == current_position:
                print("\t ---- current position")
            if t == 'SFF':
                print("\ttype: SFF instance id: {0}".format(str(instance.id)))
            elif t == 'SFI':
                print("\ttype: SFI type: {0} @ SFF {1} of Server {2}".format(str(instance.of_type), instance.sffId,
                                                                             instance.server.id))
            else:
                raise NameError(f"Something goes wrong with the path! - {path}")


class NetworkDelayEvent(PacketHoldingEvent):
    def __init__(self, delay, inner_packet: Packet, source, dest_id,
                 source_is_sff=True, dest_is_sff=True):
        sim = inner_packet.flow.sim
        super().__init__(sim.currentTime + delay, inner_packet)
        if sim.PACKET_ID_TO_DEBUG == inner_packet.id:
            print("** create network delay event of debug packet")
        self.source = source
        self.destID = dest_id
        self.source_is_sff = source_is_sff
        self.dest_is_sff = dest_is_sff
        sim.props.sff.counter_packets_put_on_wire += 1
    
    def update_packet_time_tracking(self):
        self.inner_packet.timeNetwork += self.inner_packet.get_delta_of_time_mark()
    
    # free resources of sourceSFF, and send packet to destSSF
    def process_event(self):
        sff_props = self.inner_packet.flow.sim.props.sff
        if sff_props.consider_link_capacity and self.source_is_sff and self.dest_is_sff:
            self.source.free_bw_resource_to_dest_id(self.destID, self.inner_packet)
            # check if we have to deque the next packet
            if len(self.source.outQueue[self.destID]) > 0:
                # check if available bw is enough for the next packet,
                # or if we have to wait till the next packet is done
                if (self.source.outQueue[self.destID][0].transmission_size <=
                        sff_props.linkBwRemaining[self.source.id][self.destID]):
                    self.source.put_packet_on_wire(
                        packet=self.source.outQueue[self.destID].pop(0), dest_id=self.destID)
        
        if self.inner_packet.flow.sim.PACKET_ID_TO_DEBUG == self.inner_packet.id:
            sim = self.inner_packet.flow.sim
            print(f"** @{sim.currentTime} at event {sim.current_event} " +
                  f"network delay event is done, so get time delta of network")
        self.update_packet_time_tracking()
        # do statistics for network time
        if self.source_is_sff and self.dest_is_sff:
            sff_props.allSFFs[self.destID].handle_packet_from_other_sff(self.inner_packet, self.source)
        elif self.dest_is_sff:
            sff_props.allSFFs[self.destID].handle_packet_from_sfi(self.inner_packet, self.source)
        elif self.source_is_sff:
            self.inner_packet.flow.sim.props.sfi.all_sfi[self.destID].enqueue_packet(self.inner_packet)
        else:
            if self.destID != self.source.sffId:
                raise NameError("invalid action, trying to send a packet from one SFI to another SFI (directly)")
            self.inner_packet.flow.sim.props.sfi.all_sfi[self.destID].enqueue_packet(self.inner_packet)



#!/usr/bin/env python3
# coding=utf-8

from .sff import *
from ..events import PacketHoldingEvent
from ..simulator import Sim


class SfiProcessEvent(PacketHoldingEvent):
    def __init__(self, processing_time, inner_packet: Packet, sfi: 'SFI'):
        super().__init__(inner_packet.flow.sim.currentTime + processing_time, inner_packet)
        self.sfi = sfi
    
    def update_packet_time_tracking(self):
        self.inner_packet.timeProcessing += self.inner_packet.get_delta_of_time_mark()
    
    def process_event(self):
        self.update_packet_time_tracking()
        self.sfi.finished_processing(self.inner_packet)


class SFI(object):
    @Sim.register_reset_global_fields
    class Props:
        def __init__(self):
            self.all_sfi: Dict[int, 'SFI'] = dict()
            self.next_free_id = 0
            self.latency_provider = -1
            self.processingRateOfSfType: List[int] = None
    
    # parameters: the SF type (int), the server which hosts this SFI, the SFF
    # which does scheduling of this SFI
    def __init__(self, of_type, hosting_server: 'Server', responsible_sff_id):
        assert isinstance(of_type, int)
        assert isinstance(responsible_sff_id, int)
        self.sim = hosting_server.sim
        sfi_props: SFI.Props = self.sim.props.sfi
        if sfi_props.processingRateOfSfType is None:
            raise NameError("you have to setup sf types")
        assert (len(sfi_props.processingRateOfSfType) > of_type)
        self.of_type = of_type
        self.server: 'Server' = hosting_server
        self.sffId = responsible_sff_id
        self.refreshShares = False
        self.cpuShares = 0
        self.cachedTimeToProcessAPacket = 1
        self.queue: deque = deque()
        self.free = True
        self.id = sfi_props.next_free_id
        sfi_props.all_sfi[self.id] = self
        sfi_props.next_free_id += 1
        
        self.sim.props.sff.allSFFs[responsible_sff_id].register_sfi(self)
    
    @staticmethod
    @Sim.register_stop_sim_catch_all_packets_hooks
    def callback_to_get_all_packets_from_sfi(sim: Sim):
        for sfi in sim.props.sfi.all_sfi.values():
            for p in sfi.queue:
                p.timeQueueProcessing += p.get_delta_of_time_mark()
                yield p
    
    def __repr__(self):
        return f"SFI({self.id}/sf{self.of_type}/{'f' if self.free else 'b'}/{len(self.queue)})"
    
    # setup the internal data structure, so that it supports SF types of numberOfSfTypes,
    # hence the data structure holds for sf types 0...(numberOfSfTypes-1),
    # (len(data) = numberOfSfTypes)
    @staticmethod
    def init_data_structure(sim: Sim, number_of_sf_types=1, latency_provider_sff_sfi=0):
        assert (sim.props.sfi.processingRateOfSfType is None)
        sim.props.sfi.processingRateOfSfType = [0 for _ in range(number_of_sf_types)]
        sim.props.sfi.latency_provider = latency_provider_sff_sfi
    
    @staticmethod
    def setup_sf_processing_rate_per_1s(sim: Sim, of_type, with_mu):
        """we set processing rate as number of packets per 1s,
        so a value of e.g., 22 means, per 1 share of processing from a server,
        a SFI of this type will process 22 packets per 1s
        hence, when a packet arrives and the queue of this SFI is empty,
        the packet will be done after 1000000 / (22 * CPU shares) µs"""
        of_type = int(of_type)
        sfi_props: SFI.Props = sim.props.sfi
        sff_props: SFF.Props = sim.props.sff
        if sfi_props.processingRateOfSfType is None or len(sfi_props.processingRateOfSfType) <= of_type:
            raise (
                NameError(
                    "You have to setup the data structure with the max " +
                    "number of SF types used. Run SFI.init_data_structure"))
        sfi_props.processingRateOfSfType[of_type] = with_mu
        # refresh cachedTimeToProcessAPacket
        for sffID in sff_props.allSFFs:
            if of_type in sff_props.allSFFs[sffID].SFIsPerType:
                for sfi in sff_props.allSFFs[sffID].SFIsPerType[of_type]:
                    sfi.refresh_processing_speed()
    
    # frees up all server shares this SFI has
    def free_all_server_shares(self):
        assert self.free
        self.server.availableShares += self.cpuShares
        self.cpuShares = 0
    
    # when called, this SFI tries to get as many shares from its server,
    # as considered based on the SFIs weight
    # this method updates the shares only, if the SFI is not in processing mode.
    # if this method is called during processing mode, we set a flag for updating
    # the shares when the processing is done.
    # otherwise, we might free resources,
    # which are still in usage. Hence, some shares are used simultaneously
    def refresh_server_shares(self):
        if not self.free:
            self.refreshShares = True
            return
        self.refreshShares = False
        # weight is in the range of [0,1000]
        # we stick to int arithmetic, for performance reasons
        w = self.server.sfiWeights[self]
        shares_target = int(self.server.processing_cap * w /
                            self.sim.SERVER_CPU_SHARE_GRANULARITY)
        # if Sim.DEBUG:
        #     print(
        #         "* targetShares is {3} based on weight {0} processing_cap {1} granularity {2}".format(
        #             str(w), str(
        #                 self.server.processing_cap), str(
        #                 Sim.SERVER_CPU_SHARE_GRANULARITY), str(shares_target)))
        
        if self.server.cpu_policy != ServerCpuPolicy.one_at_a_time:
            assert (shares_target > 0)
        
        # if Sim.DEBUG:
        #     print("{0} has a weight of {1}, holds {2} shares and wants {3} shares".format(
        #         self, str(w), str(self.cpuShares), str(shares_target)))
        if self.cpuShares > shares_target:
            # we have to free some shares
            to_give_free = self.cpuShares - shares_target
            assert to_give_free > 0
            self.server.availableShares += to_give_free
            self.cpuShares = shares_target
        else:
            # we are allowed to take more shares,
            needed_shares = shares_target - self.cpuShares
            if self.server.availableShares >= needed_shares:
                # there are enough free shares, so take them
                self.server.availableShares -= needed_shares
                self.cpuShares = shares_target
            else:
                # some shares are missing, but take as many as possible
                self.cpuShares += self.server.availableShares
                self.server.availableShares = 0
                # reschedule this method to get more shares later
                if self.sim.DEBUG:
                    print("{0} shares missing, i have only {1}".format(
                        self, str(self.cpuShares)))
                self.refreshShares = True
        if self.sim.DEBUG:
            print("{0} holds now {1} shares, server has {2} left".format(
                self, str(self.cpuShares), str(self.server.availableShares)
            ))
        self.refresh_processing_speed()
    
    # returns the expected waiting time when a packet will be scheduled on this sfi
    # however, the returned value might be strongly wrong, depending on the server cpu policy
    def get_expected_waiting_time(self) -> int:
        if self.server.cpu_policy == ServerCpuPolicy.one_at_a_time:
            return len(self.queue) * int(
                1000000 / (self.server.processing_cap * self.sim.props.sfi.processingRateOfSfType[self.of_type]))
        else:
            return self.cachedTimeToProcessAPacket * len(self.queue)
    
    def get_expected_processing_rate(self) -> float:
        if self.server.cpu_policy == ServerCpuPolicy.one_at_a_time:
            return self.server.processing_cap * self.sim.props.sfi.processingRateOfSfType[self.of_type]
        else:
            return 1000000.0 / self.get_expected_processing_time()
    
    def get_expected_processing_time(self) -> int:
        if self.server.cpu_policy == ServerCpuPolicy.one_at_a_time:
            delay = int(1000000 // self.get_expected_processing_rate())
            if delay == 0:
                raise NameError("the speed of this sfi is too high, so that the time for processing is less than 1."
                                "Server capacity: {0}, µ of sfi:{1}".format(self.server.processing_cap,
                                                                            self.sim.props.sfi.processingRateOfSfType[
                                                                                self.of_type]))
            return delay
        else:
            return self.cachedTimeToProcessAPacket
    
    # updates the time required for each packet based on the amount of shares
    # the sfi has
    def refresh_processing_speed(self):
        # if Sim.DEBUG:
        #     print(
        #         "{0} has {1} shares and updates the time required for a packet".format(
        #             self, self.cpuShares))
        
        assert (self.sim.props.sfi.processingRateOfSfType[self.of_type] > 0)
        
        # if cpu policy of the server is one at a time, we might have 0 cpu shares
        if self.server.cpu_policy == ServerCpuPolicy.one_at_a_time:
            if self.cpuShares == 0:
                return
        else:
            # otherwise check that we hat at least some shares
            assert (self.cpuShares > 0)
        
        self.cachedTimeToProcessAPacket = int(
            1000000 / (self.sim.props.sfi.processingRateOfSfType[self.of_type] * self.cpuShares))
        # print(f"sfi processing time for type {self.of_type} is {self.cachedTimeToProcessAPacket}")
    
    # notify the sfi that it is allowed for starting processing queued events
    def notify_for_processing(self):
        assert self.free
        assert len(self.queue) > 0
        assert self.server.ask_for_processing(self)
        self.free = False
        self.internal_schedule_event(self.queue.popleft())
    
    def finished_processing(self, packet: Packet):
        # do we need to refresh our server shares we reserve?
        if self.refreshShares:
            self.free = True
            self.refresh_server_shares()
            self.free = False
        
        # handle next hop of packet
        if self.sim.DEBUG:
            print(f"processing done for packet {packet.id} at sfi {self.id}")
            Packet.debug_print_path(packet.fullPath)
        
        if self.server.cpu_policy == ServerCpuPolicy.one_at_a_time:
            self.free = True
            self.server.sfi_finishes_processing(self)
        else:
            if len(self.queue) > 0:
                self.internal_schedule_event(self.queue.popleft())
            else:
                self.free = True
        
        sff_props: SFF.Props = self.sim.props.sff
        sff_props.allSFFs[self.sffId].sfi_finishes_processing_of_packet(self, packet)
        
        next_hop_type, next_hop = packet.fullPath[packet.pathPosition]
        packet.pathPosition += 1
        # we also have to add 1 to the position with respect to the sfc
        
        # is this the last stop of this sfc?
        if Flow.get_sf_and_eoc_of_packet(packet)[1]:
            packet.processing_done = True
        
        packet.sfc_position += 1
        
        packet.mark_time()
        sfi_props: SFI.Props = self.sim.props.sfi
        sff_props: SFF.Props = self.sim.props.sff
        
        delay = next(sff_props.latencyProvider[sfi_props.latency_provider])
        
        if next_hop_type == SFF.__name__:
            # we send this packet to a SFF
            # this should be the SFF to which this SFI belongs to
            assert (self.sffId == next_hop.id)
            # since this does not take any latency, we simply call handle
            # packet
            if self.sim.PACKET_ID_TO_DEBUG == packet.id:
                print(f'** packet processed at sfi {self.id}, send it back to sff')
            self.sim.schedule_event(NetworkDelayEvent(delay=delay,
                                                      inner_packet=packet,
                                                      source=self,
                                                      dest_id=self.sffId,
                                                      source_is_sff=False,
                                                      dest_is_sff=True))
        
        elif next_hop_type == SFI.__name__:
            # we send this packet to the next SFI
            # this SFI should have the same SFF as I have, otherwise this is not
            # allowed
            if self.sim.PACKET_ID_TO_DEBUG == packet.id:
                print(f'** packet processed at sfi {self.id}, send it to next sfi {next_hop.id}')
            self.sim.schedule_event(NetworkDelayEvent(delay=delay,
                                                      inner_packet=packet,
                                                      source=self,
                                                      dest_id=self.sffId,
                                                      source_is_sff=False,
                                                      dest_is_sff=False))
        else:
            NameError("unknown next hop")
    
    def internal_schedule_event(self, packet: Packet):
        assert not self.free
        
        # check if packet is already outdated
        while packet.flow.qosMaxDelay < (self.sim.currentTime - packet.time_ingress + self.cachedTimeToProcessAPacket):
            if self.sim.DEBUG:
                print("drop packet because of timeout")
            
            # add up queue time of this packet
            packet.timeQueueProcessing += packet.get_delta_of_time_mark()
            
            # we drop a packet, so we have to be in free state when calling this
            self.free = True
            packet.drop_timed_out(self)
            # is there any other packet and are we still in free state?
            # in case of one at a time, we have to check whether we are allowed to proceed
            if len(self.queue) > 0 and self.free and \
                    (self.server.cpu_policy != ServerCpuPolicy.one_at_a_time or self.server.ask_for_processing(self)):
                # yes, so set us a busy
                self.free = False
                packet = self.queue.popleft()
            else:
                self.free = True
                # we have to tell the server that we are done
                if self.server.cpu_policy == ServerCpuPolicy.one_at_a_time:
                    self.server.sfi_finishes_processing(self)
                return
        
        if self.sim.DEBUG:
            print(f".. start processing a packet {packet.id} at sfi {self.id} @{self.sim.currentTime}")
            
            if self.server.cpu_policy == ServerCpuPolicy.one_at_a_time:
                # now we are in isolated mode, sanity check
                running_server = set()
                for sfi_id in self.sim.props.sfi.all_sfi:
                    sfi = self.sim.props.sfi.all_sfi[sfi_id]
                    server = sfi.server
                    if not sfi.free:
                        assert server not in running_server
                        running_server.add(server)
                
                assert self.server.availableShares == 0
                assert self.cpuShares == self.server.processing_cap
        
        packet.set_callback_when_dropped(None)
        # add up queue time of this packet
        packet.timeQueueProcessing += packet.get_delta_of_time_mark()
        
        # process packet
        packet.mark_time()
        self.sim.schedule_event(
            SfiProcessEvent(
                processing_time=self.cachedTimeToProcessAPacket,
                inner_packet=packet,
                sfi=self))
    
    def enqueue_packet(self, packet: Packet):
        if self.sim.TRACE_PACKET_PATH:
            packet.visitedHops.append(self)
        
        if packet.id == self.sim.PACKET_ID_TO_DEBUG:
            print(f"** receive debug packet at sfi {self}")
        
        # remember current time so that we track queue time for this packet
        packet.mark_time()
        
        if self.free:
            if self.server.ask_for_processing(self):
                # server is not free, so queue this packet
                
                self.free = False
                self.internal_schedule_event(packet)
            else:
                self.queue.append(packet)
        
        else:
            self.queue.append(packet)



#!/usr/bin/env python3
# coding=utf-8
from typing import Set

from .sfi import *
from ..events import BaseEvent
from ..simulator import Sim


class ServerCpuShareEvent(BaseEvent):
    def __init__(self, interval, server):
        super().__init__(server.sim.currentTime + interval)
        self.ignoreWhenFinished = True
        self.server = server
        self.interval = interval
    
    def process_event(self):
        # reschedule next update
        self.server.sim.schedule_event(ServerCpuShareEvent(
            interval=self.interval, server=self.server))
        # trigger cpu share update
        self.server.update_dynamic_cpu_weights()


class Server(object):
    @Sim.register_reset_global_fields
    class Props:
        def __init__(self):
            self.lastId: int = 0
            self.all_servers: List['Server'] = []
    
    def __init__(self, sim: Sim, processing_cap, cpu_policy: ServerCpuPolicy):
        self.SFIs = []
        self.sim = sim
        self.SFF_ids: Set[int] = set()
        sim.props.server.lastId += 1
        self.id = sim.props.server.lastId
        self.processing_cap = int(processing_cap)
        self.availableShares = self.processing_cap
        self.sfiWeights = {}
        self.cpu_policy = cpu_policy
        
        self.stats_idle_time = 0
        self.stats_last_time_idle = 0
        
        if self.cpu_policy == ServerCpuPolicy.dynamic:
            sim.schedule_event(ServerCpuShareEvent(
                interval=sim.SERVER_CPU_POLICY_DYNAMIC_INTERVAL, server=self))
        
        sim.props.server.all_servers.append(self)
    
    def __repr__(self):
        return f"Server({str(self.id)}/{'f' if self.is_free() else 'b'})"
    
    # checks if all hosted sfis are in idle mode
    def is_free(self):
        for sfi in self.SFIs:
            if not sfi.free:
                return False
        
        return True
    
    # tests whether the asking sfi is allowed to go in processing mode
    # if so, we update the cpu shares, update all affected sfis and return true
    def ask_for_processing(self, asking_sfi: SFI):
        is_free = self.is_free()
        
        if self.sim.DEBUG:
            print(
                f"server {self.id} is asked for giving processing token to a sfi {asking_sfi.id}. My answer: {is_free}")
        
        if is_free:
            self.stats_idle_time += self.sim.currentTime - self.stats_last_time_idle
            self.stats_last_time_idle = self.sim.currentTime
        
        if self.cpu_policy != ServerCpuPolicy.one_at_a_time:
            return True
        else:
            if is_free:
                for sfi in self.sfiWeights:
                    if asking_sfi == sfi:
                        self.sfiWeights[sfi] = self.sim.SERVER_CPU_SHARE_GRANULARITY
                    else:
                        self.sfiWeights[sfi] = 0
                
                self.notify_sfi_to_recalculate_shares()
                # we have to inform the asking sfi two times,
                # since it might happen that in the previous call this sfi
                # was triggered before a different sfi freed his shares
                asking_sfi.refresh_server_shares()
                return True
            else:
                return False
    
    def sfi_finishes_processing(self, sfi: SFI):
        self.stats_last_time_idle = self.sim.currentTime
        # if we are in cpu policy one at a time, we check if we have to inform some other sfi to start processing
        
        if self.cpu_policy == ServerCpuPolicy.one_at_a_time:
            if self.is_free():
                shuffled_sfi = [i for i in self.SFIs]
                self.sim.random.shuffle(shuffled_sfi)
                for sfi in shuffled_sfi:
                    if len(sfi.queue) > 0:
                        sfi.notify_for_processing()
                        return
    
    def notify_sfi_to_recalculate_shares(self):
        for sfi in self.SFIs:
            sfi.refresh_server_shares()
    
    def update_dynamic_cpu_weights(self):
        if len(self.SFIs) == 0:
            if self.sim.DEBUG:
                print("WARN - server {0} has no SFIs".format(self))
            return
        # get the total queue length among all SFIs
        total_queue_length = 0
        for sfi in self.SFIs:
            total_queue_length += len(sfi.queue)
        denominator = total_queue_length + len(self.SFIs)
        # we have to make sure, that each sfiWeights is at least as big enough,
        # so that each SFI gets at least one CPU share (prevent starvation!)
        weight_for_one_share = int(
            self.sim.SERVER_CPU_SHARE_GRANULARITY / self.processing_cap) + 1
        weights_free_to_assign = self.sim.SERVER_CPU_SHARE_GRANULARITY - \
                                 weight_for_one_share * len(self.SFIs)
        
        weight_remainder = self.sim.SERVER_CPU_SHARE_GRANULARITY
        for sfi in self.sfiWeights:
            enumerator = (len(sfi.queue) + 1)
            self.sfiWeights[sfi] = weight_for_one_share + \
                                   int(weights_free_to_assign * enumerator / denominator)
            weight_remainder -= self.sfiWeights[sfi]
            if self.sim.DEBUG:
                print(" * {0} has an enume of {1} and gets weight {2}".format(
                    sfi, str(enumerator), str(self.sfiWeights[sfi])))
        while weight_remainder > 0:
            for sfi in self.SFIs:
                if weight_remainder > 0:
                    self.sfiWeights[sfi] += 1
                    weight_remainder -= 1
        
        self.notify_sfi_to_recalculate_shares()
    
    def add_sfi(self, of_type, with_sff_id):
        self.SFF_ids.add(with_sff_id)
        sfi = SFI(of_type, self, with_sff_id)
        
        self.SFIs.append(sfi)
        if len(self.SFIs) > self.processing_cap and self.cpu_policy == ServerCpuPolicy.static:
            raise NameError("more SFIs running ({0}) on this server ".format(str(len(
                self.SFIs))) + "than cpuShares ({0}) available".format(str(self.processing_cap)))
        
        self.sfiWeights[sfi] = 0
        if self.cpu_policy == ServerCpuPolicy.static:
            eq = int(self.sim.SERVER_CPU_SHARE_GRANULARITY / len(self.sfiWeights))
            if self.sim.DEBUG:
                print(
                    "{0} splits his shares ({1}) according to the static policy, each weight is {2}".format(
                        self, str(
                            self.processing_cap), str(eq)))
            assert (eq > 0)
            for sfi in self.sfiWeights:
                self.sfiWeights[sfi] = eq
                # in case of static policy, we are allowed to reset all shares,
                # at this point in time, because no SFI runs a packet currently
                sfi.free_all_server_shares()
            self.notify_sfi_to_recalculate_shares()
        elif self.cpu_policy == ServerCpuPolicy.dynamic:
            self.update_dynamic_cpu_weights()



#!/usr/bin/env python3
# coding=utf-8
import re
import time
from typing import List

from .model.server import Server
from .model.sff import SFF
from .model.sfi import SFI
from .model.core import Flow, Packet
from .simulator import Sim
from .events import BaseEvent


class StatisticsPollingEvent(BaseEvent):
    def __init__(self, interval, stats_writer: 'SimStatsPoller', first_event=False):
        super().__init__(stats_writer.sim.currentTime + (0 if first_event else interval))
        self.ignoreWhenFinished = True
        self.stats_writer = stats_writer
        self.interval = interval
    
    def process_event(self):
        # reschedule next update
        self.stats_writer.sim.schedule_event(
            StatisticsPollingEvent(interval=self.interval, stats_writer=self.stats_writer))
        self.stats_writer.poll_statistics()


class OverviewStatisticsEvent(BaseEvent):
    
    def __init__(self, list_of_snapshots: List[int], sim: Sim):
        super().__init__(list_of_snapshots.pop(0))
        self.ignoreWhenFinished = True
        self.sim = sim
        self.list_of_snapshots = list_of_snapshots
    
    def process_event(self):
        # reschedule next update
        stat_props: SimStats.Props = self.sim.props.sim_stats
        # if overview statistics is not active, ignore this event
        if stat_props.overviewStats is None:
            return
        
        if len(self.list_of_snapshots) > 0:
            self.sim.schedule_event(
                OverviewStatisticsEvent(list_of_snapshots=self.list_of_snapshots, sim=self.sim))
        SimStats.do_overview_statistics_snapshot(self.sim)


class SimStatsWriter(object):
    
    def __init__(self, sim: Sim, filepath):
        self.filepath = filepath
        self.sim = sim
        self.entries = []
        self.workload_is_over = False
        
        sim.props.sim_stats.allStatsWriter.append(self)
        if sim.DEBUG:
            print("* genStatWriter for {0} called".format(self.filepath))
    
    def init_file(self):
        with open(self.filepath, "w") as f:
            f.write(self.header_row())
    
    def header_row(self):
        raise NameError("must be implemented by subclass")
    
    def flush(self):
        if self.sim.DEBUG:
            print(
                "* flush {0} elements to {1}".format(str(len(self.entries)), self.filepath))
        with open(self.filepath, "a") as f:
            for e in self.entries:
                f.write("\n")
                f.write(e)
        self.entries = []
    
    def inform_workload_is_over(self):
        self.workload_is_over = True


class SimStatsRowWriter(SimStatsWriter):
    
    def __init__(self, sim: Sim, filepath, columns: list, data_types: list):
        super().__init__(sim, filepath)
        self.columns = columns
        self.data_types = data_types
        self.init_file()
    
    def header_row(self):
        return self.sim.props.sim_stats.SEPARATOR.join(self.columns)
    
    def add_entry(self, *vals):
        s = []
        assert (len(vals) == len(self.columns))
        for i, v in enumerate(vals):
            s.append(str(self.data_types[i](v)))
        self.entries.append(self.sim.props.sim_stats.SEPARATOR.join(s))
        if len(self.entries) > self.sim.props.sim_stats.FLUSH_ENTRIES:
            self.flush()


class SimStatsKvWriter(SimStatsWriter):
    
    def __init__(self, sim: Sim, filepath, data_type: type = float):
        super().__init__(sim, filepath)
        self.init_file()
        self.datatype = data_type
    
    def header_row(self):
        return self.sim.props.sim_stats.SEPARATOR.join(["time", "key", "value"])
    
    def add_kv_entry(self, key, value):
        self.entries.append(self.sim.props.sim_stats.SEPARATOR.join([
            str(self.sim.currentTime), str(key), str(self.datatype(value))]))
        if len(self.entries) > self.sim.props.sim_stats.FLUSH_ENTRIES:
            self.flush()


class SimStatsCdfWriter(SimStatsWriter):
    
    def __init__(self, sim: Sim, filepath, buckets: int, per_group: bool):
        super().__init__(sim, filepath)
        self.init_file()
        self.buckets = buckets
        self.bucket_holder = dict()
        self.per_group = per_group
    
    def prepare_flush(self):
        sep = self.sim.props.sim_stats.SEPARATOR
        for key in self.bucket_holder:
            s_key = str(key)
            for group in self.bucket_holder[key]:
                s_group = s_key + sep + str(group)
                for bucket, number in enumerate(self.bucket_holder[key][group]):
                    self.entries.append(sep.join([s_group, str(bucket), str(number)]))
        
        self.bucket_holder = None  # intentionally, we set bucket_holder to None, so that after flushing this guy,
        # there is no reason to 1. flush it again, 2. add some more data points
    
    def header_row(self):
        return self.sim.props.sim_stats.SEPARATOR.join(["key", "group", "bucket", "count"])
    
    def add_data_point(self, key, group, value: float):
        if value > 1 or value < 0:
            raise NameError(f"received invalid value ({value}) for cdf - expect to receive normalized valued [0,1]")
        
        if key not in self.bucket_holder:
            self.bucket_holder[key] = dict()
        
        if not self.per_group:
            group = "all"
        
        if group not in self.bucket_holder[key]:
            self.bucket_holder[key][group] = [0 for _ in range(self.buckets)]
        
        bucket = self.buckets - 1 if value == 1 else int(value * self.buckets)
        self.bucket_holder[key][group][bucket] += 1


class SimStatsPoller(SimStatsRowWriter):
    
    def __init__(self, sim: Sim, filepath, interval):
        super().__init__(sim, filepath, ['time', 'stat',
                                         'key', 'value'], [int, str, str, float])
        self.interval = interval
        self.sim.schedule_event(StatisticsPollingEvent(
            interval=interval, stats_writer=self, first_event=True))
        self.allServers = []
        self.SFIs = dict()
        self.SFFs = []
        self.overview = False
        self.overview_last_value_cache = dict()
    
    def poll_statistics(self):
        flow_props: Flow.Props = self.sim.props.flow
        packet_props: Packet.Props = self.sim.props.packet
        server_props: Server.Props = self.sim.props.server
        sff_props: SFF.Props = self.sim.props.sff
        sfi_props: SFI.Props = self.sim.props.sfi
        
        # simulator events length
        if self.sim.DEBUG:
            self.add_entry(self.sim.currentTime, "simulator", "queue", self.sim.event_list.len())
        
        if self.overview:
            idle_time = 0
            for s in self.sim.props.server.all_servers:
                idle_time += s.stats_idle_time
                if s.is_free():
                    idle_time += self.sim.currentTime - s.stats_last_time_idle
            
            if len(self.overview_last_value_cache) == 0:
                self.overview_last_value_cache["packets_done"] = packet_props.statsPacketsSuccessfulProcessed
                self.overview_last_value_cache["time"] = self.sim.currentTime
                self.overview_last_value_cache["idle_time"] = idle_time
                self.overview_last_value_cache["packet_ingress"] = packet_props.counter_ingress_packets
                self.overview_last_value_cache["packet_rejected"] = packet_props.statsPacketsRejectedSchedule
                self.overview_last_value_cache["packet_timeout"] = packet_props.statsPacketsRejectedProcessingDelay
                self.overview_last_value_cache["packet_successful"] = packet_props.statsPacketsSuccessfulProcessed
            else:
                rates = {
                    "goodput": (packet_props.statsPacketsSuccessfulProcessed - self.overview_last_value_cache[
                        "packets_done"]) /
                               ((self.sim.currentTime - self.overview_last_value_cache["time"]) / 1000000),
                    "packet_rejected_rate": (packet_props.statsPacketsRejectedSchedule - self.overview_last_value_cache[
                        "packet_rejected"]) /
                                            ((self.sim.currentTime - self.overview_last_value_cache["time"]) / 1000000),
                    
                    "packet_ingress_rate": (packet_props.counter_ingress_packets - self.overview_last_value_cache[
                        "packet_ingress"]) /
                                           ((self.sim.currentTime - self.overview_last_value_cache["time"]) / 1000000),
                    
                    "packet_timeout_rate": (packet_props.statsPacketsRejectedProcessingDelay -
                                            self.overview_last_value_cache[
                                                "packet_timeout"]) /
                                           ((self.sim.currentTime - self.overview_last_value_cache["time"]) / 1000000),
                    
                    "packet_success_rate": (packet_props.statsPacketsSuccessfulProcessed -
                                            self.overview_last_value_cache[
                                                "packet_successful"]) /
                                           ((self.sim.currentTime - self.overview_last_value_cache["time"]) / 1000000),
                    
                    "server_idle_ratio": (idle_time - self.overview_last_value_cache["idle_time"]) /
                                         (self.sim.currentTime - self.overview_last_value_cache["time"]) /
                                         len(server_props.all_servers),
                }
                for k, v in rates.items():
                    self.add_entry(self.sim.currentTime, "overview", k, v)
            
            self.add_entry(self.sim.currentTime, "overview", "packet_in_system", packet_props.counter_packet_in_system)
            self.add_entry(self.sim.currentTime, "overview", "packet_total_count", packet_props.statsPacketsTotalCount)
            self.add_entry(self.sim.currentTime, "overview", "packet_rejected",
                           packet_props.statsPacketsRejectedSchedule)
            self.add_entry(self.sim.currentTime, "overview", "packet_timeout",
                           packet_props.statsPacketsRejectedProcessingDelay)
            self.add_entry(self.sim.currentTime, "overview", "packet_successful",
                           packet_props.statsPacketsSuccessfulProcessed)
            self.add_entry(self.sim.currentTime, "overview", "packet_total_delay", packet_props.statsSumDelay)
            if packet_props.statsPacketsSuccessfulProcessed > 0:
                self.add_entry(self.sim.currentTime, "overview", "packet_delivered_delay_avg",
                               packet_props.statsRatiosQos / packet_props.statsPacketsSuccessfulProcessed)
            else:
                self.add_entry(self.sim.currentTime, "overview", "packet_delivered_delay_avg", 1)
            
            self.add_entry(self.sim.currentTime, "overview", "sff_queue",
                           sum([sff.get_number_of_queued_packets() for sff in sff_props.allSFFs.values()]))
            self.add_entry(self.sim.currentTime, "overview", "sfi_queue",
                           sum([len(sfi.queue) for sfi in sfi_props.all_sfi.values()]))
            
            self.overview_last_value_cache["packets_done"] = packet_props.statsPacketsSuccessfulProcessed
            self.overview_last_value_cache["time"] = self.sim.currentTime
            self.overview_last_value_cache["idle_time"] = idle_time
            self.overview_last_value_cache["packet_ingress"] = packet_props.counter_ingress_packets
            self.overview_last_value_cache["packet_rejected"] = packet_props.statsPacketsRejectedSchedule
            self.overview_last_value_cache["packet_timeout"] = packet_props.statsPacketsRejectedProcessingDelay
            self.overview_last_value_cache["packet_successful"] = packet_props.statsPacketsSuccessfulProcessed
        
        # sfi statistics
        for sfiType in self.SFIs:
            for sfi in self.SFIs[sfiType]:
                self.add_entry(self.sim.currentTime, "sfiQueue",
                               sfiType, len(sfi.queue))
                self.add_entry(self.sim.currentTime, "sfiState",
                               sfiType, int(sfi.free))
                self.add_entry(self.sim.currentTime, "sfiCpuShares",
                               sfiType, sfi.cpuShares)
        
        # server statistics
        free_servers = 0
        for server in self.allServers:
            is_free = server.is_free()
            if is_free:
                free_servers += 1
            # self.add_entry(self.sim.currentTime, "server",
            #                "availableShares", server.availableShares)
            self.add_entry(self.sim.currentTime, "server", "is_free", 1 if is_free else 0)
        
        # sff Statistics
        if len(self.SFFs) > 0:
            # self.add_entry(self.sim.currentTime, "sff", "networkQueue",
            #                sum([len(x) for x in sff.outQueue.values()]))
            
            queue_per_sf = {k: 0 for k in range(len(sfi_props.processingRateOfSfType))}
            for sff in self.SFFs:
                if sff_props.consider_link_capacity:
                    self.add_entry(self.sim.currentTime, "sff", "networkQueue",
                                   sum([len(x) for x in sff.outQueue.values()]))
                if sff.scheduler.requires_queues_per_class():
                    for q in sff.packet_queue_per_class:
                        queue_per_sf[flow_props.sfc_class_to_sf[q][0]] += len(sff.packet_queue_per_class[q])
                else:
                    queue_per_sf[0] += len(sff.packet_queue)
                
                self.add_entry(self.sim.currentTime, "sff_queue", sff.id, sff.get_number_of_queued_packets())
            
            for sf in queue_per_sf:
                self.add_entry(self.sim.currentTime, "sf_queue", sf, queue_per_sf[sf])


class SimStats(object):
    @Sim.register_reset_global_fields
    class Props:
        def __init__(self):
            self.exp_id = None
            self.flowIDs: set = set()
            self.flowStats: SimStatsRowWriter = None
            self.allStatsWriter: List[SimStatsWriter] = []
            self.SEPARATOR: str = ','
            self.FLUSH_ENTRIES: int = 10000
            self.debugStats: SimStatsKvWriter = None
            self.packetStats: SimStatsRowWriter = None
            self.packetCdfStats: SimStatsCdfWriter = None
            self.pollStatistics: SimStatsPoller = None
            self.exp_stats: SimStatsKvWriter = None
            self.serverStats: SimStatsRowWriter = None
            self.overviewStats: SimStatsKvWriter = None
            self.workloadStats: SimStatsRowWriter = None
    
    @staticmethod
    def activate_packet_statistics(sim: Sim):
        if sim.props.sim_stats.exp_id is None:
            raise NameError("you have to activate statistics ... SimStats.activate")
        filepath = "stats_{0}_packets.csv".format(sim.props.sim_stats.exp_id)
        sim.props.sim_stats.packetStats = SimStatsRowWriter(
            sim=sim,
            filepath=filepath,
            columns=[
                "id",
                "flow_id",
                "ingress_time",
                "status",
                "time_total",
                "time_processing",
                "time_processing_queue",
                "time_network",
                "time_network_queue",
                "time_queue_scheduling",
                "real_time_scheduling",
                "seen_by_schedulers",
                "ingress_egress_delay"],
            data_types=[
                int,
                int,
                int,
                str,
                int,
                int,
                int,
                int,
                int,
                int,
                float,
                int,
                int])
    
    @staticmethod
    def activate(sim: Sim, exp_id: str, configuration: dict = None):
        stat_props: SimStats.Props = sim.props.sim_stats
        assert (stat_props.exp_id is None)
        if not re.match('^[a-zA-Z0-9\-_]+$', exp_id):
            raise NameError("experiment ID should be alphanumeric")
        stat_props.exp_id = exp_id
        filepath = "stats_{0}.csv".format(stat_props.exp_id)
        stat_props.exp_stats = SimStatsKvWriter(sim=sim, filepath=filepath, data_type=str)
        stat_props.exp_stats.add_kv_entry(key="exp_id", value=exp_id)
        if configuration is not None:
            for k in configuration:
                stat_props.exp_stats.add_kv_entry(key=k, value=f"\"{configuration[k]}\"")
    
    @staticmethod
    def activate_workload_statistics(sim: Sim):
        stat_props: SimStats.Props = sim.props.sim_stats
        if stat_props.exp_id is None:
            raise NameError("you have to activate statistics ... SimStats.activate")
        filepath = "stats_{0}_workload.csv".format(stat_props.exp_id)
        stat_props.workloadStats = SimStatsRowWriter(sim=sim, filepath=filepath,
                                                     columns=["time", "flow_id", "chain", "ingress",
                                                              "egress", "qos_delay"],
                                                     data_types=[int, int, str, int,
                                                                 int, int])
    
    @staticmethod
    def activate_flow_statistics(sim: Sim):
        stat_props: SimStats.Props = sim.props.sim_stats
        if stat_props.exp_id is None:
            raise NameError("you have to activate statistics ... SimStats.activate")
        filepath = "stats_{0}_flows.csv".format(stat_props.exp_id)
        stat_props.flowStats = SimStatsRowWriter(sim=sim, filepath=filepath,
                                                 columns=["flow_id", "chain",
                                                          "egress", "qos_delay"],
                                                 data_types=[int, str, int, int])
    
    @staticmethod
    def activate_packet_cdf_statistics(sim: Sim, cdf_buckets: int, per_group: bool = True):
        stat_props: SimStats.Props = sim.props.sim_stats
        if stat_props.exp_id is None:
            raise NameError("you have to activate statistics ... SimStats.activate")
        filepath = "stats_{0}_packet_cdf.csv".format(stat_props.exp_id)
        stat_props.packetCdfStats = SimStatsCdfWriter(sim=sim, filepath=filepath,
                                                      buckets=cdf_buckets,
                                                      per_group=per_group)
    
    @staticmethod
    def activate_server_statistics(sim: Sim):
        stat_props: SimStats.Props = sim.props.sim_stats
        if stat_props.exp_id is None:
            raise NameError("you have to activate statistics ... SimStats.activate")
        filepath = "stats_{0}_server.csv".format(stat_props.exp_id)
        stat_props.serverStats = SimStatsRowWriter(sim=sim, filepath=filepath,
                                                   columns=["server_id", "capacity", "number_of_sfi",
                                                            "number_of_sff",
                                                            "idle_time"],
                                                   data_types=[int, int, int, int, int])
    
    @staticmethod
    def activate_overview_statistics(sim: Sim, optional_time_snapshots: List[int] = None):
        stat_props: SimStats.Props = sim.props.sim_stats
        if stat_props.exp_id is None:
            raise NameError("you have to activate statistics ... SimStats.activate")
        filepath = "stats_{0}_overview.csv".format(stat_props.exp_id)
        stat_props.overviewStats = SimStatsKvWriter(sim=sim, filepath=filepath, data_type=float)
        if optional_time_snapshots is not None:
            sim.schedule_event(OverviewStatisticsEvent(list_of_snapshots=optional_time_snapshots, sim=sim))
    
    @staticmethod
    def activate_debug_statistics(sim: Sim, filepath):
        stat_props: SimStats.Props = sim.props.sim_stats
        stat_props.debugStats = SimStatsKvWriter(sim=sim, filepath=filepath)
    
    @staticmethod
    def activate_polling_statistics(sim: Sim, interval):
        stat_props: SimStats.Props = sim.props.sim_stats
        if stat_props.exp_id is None:
            raise NameError("you have to activate statistics ... SimStats.activate")
        filepath = "stats_{0}_polling.csv".format(stat_props.exp_id)
        stat_props.pollStatistics = SimStatsPoller(sim=sim,
                                                   filepath=filepath, interval=interval)
    
    @staticmethod
    def activate_polling_sff_statistics(sim: Sim):
        stat_props: SimStats.Props = sim.props.sim_stats
        assert (not (sim.props.sim_stats.pollStatistics is None))
        stat_props.pollStatistics.SFFs = []
        if len(sim.props.sff.allSFFs) == 0:
            raise NameError("you should call this method after setting up your experiment")
        for sff in sim.props.sff.allSFFs.values():
            stat_props.pollStatistics.SFFs.append(sff)
    
    @staticmethod
    def activate_polling_sfi_statistics(sim: Sim):
        stat_props: SimStats.Props = sim.props.sim_stats
        assert (not (stat_props.pollStatistics is None))
        stat_props.pollStatistics.SFIs = dict()
        if len(sim.props.sff.allSFFs) == 0:
            raise NameError("you should call this method after setting up your experiment")
        for sff in sim.props.sff.allSFFs.values():
            for sfiArray in sff.SFIsPerType.values():
                for sfi in sfiArray:
                    stat_props.pollStatistics.SFIs.setdefault(
                        sfi.of_type, []).append(sfi)
    
    @staticmethod
    def activate_polling_server_statistics(sim: Sim):
        stat_props: SimStats.Props = sim.props.sim_stats
        assert (not (stat_props.pollStatistics is None))
        stat_props.pollStatistics.allServers = []
        if len(sim.props.sff.allSFFs) == 0:
            raise NameError("you should call this method after setting up your experiment")
        for sff in sim.props.sff.allSFFs.values():
            for sfiArray in sff.SFIsPerType.values():
                for sfi in sfiArray:
                    stat_props.pollStatistics.allServers.append(sfi.server)
    
    @staticmethod
    def activate_polling_overview_statistics(sim: Sim):
        stat_props: SimStats.Props = sim.props.sim_stats
        assert (not (stat_props.pollStatistics is None))
        stat_props.pollStatistics.overview = True
    
    @staticmethod
    @Packet.register_tear_down
    def callback_packet_teardown(packet: Packet):
        sim = packet.flow.sim
        stat_props: SimStats.Props = sim.props.sim_stats
        if not (stat_props.packetStats is None):
            stat_props.packetStats.add_entry(
                packet.id,
                packet.flow.id,
                packet.time_ingress,
                packet.final_state,
                packet.delay,
                packet.timeProcessing,
                packet.timeQueueProcessing,
                packet.timeNetwork,
                packet.timeQueueNetwork,
                packet.timeQueueScheduling,
                packet.realTimeScheduling,
                packet.seenByScheduler,
                SFF.get_multi_hop_latency_for(sim, packet.ingress_sff_id, packet.flow.desiredEgressSSFid))
        
        expected_delay = packet.timeProcessing + packet.timeNetwork + \
                         packet.timeQueueScheduling + packet.timeQueueProcessing + packet.timeQueueNetwork
        if not packet.delay == expected_delay:
            print(f"\n timeProcessing:{packet.timeProcessing} timeNetwork:{packet.timeNetwork} "
                  f"timeQueueNetwork:{packet.timeQueueNetwork} timeQueueProcessing:{packet.timeQueueProcessing} "
                  f"timeQueueScheduling:{packet.timeQueueScheduling} \n.. sum: {expected_delay} "
                  f"ingress time:{packet.time_ingress} now:{sim.currentTime} -> delay:{packet.delay}")
            raise NameError("something goes wrong with this packet!")
        
        if not (stat_props.packetCdfStats is None):
            # latency (service quality)
            value = packet.delay / packet.flow.qosMaxDelay if (packet.final_state == "done") else 1
            stat_props.packetCdfStats.add_data_point(key="latency",
                                                     group=packet.flow.sfc_identifier,
                                                     value=value)
        
        if not (stat_props.flowStats is None):
            if not (packet.flow.id in stat_props.flowIDs):
                stat_props.flowIDs.add(packet.flow.id)
                stat_props.flowStats.add_entry(packet.flow.id,
                                               '-'.join(map(str,
                                                            packet.flow.sfTypeChain)),
                                               packet.flow.desiredEgressSSFid,
                                               packet.flow.qosMaxDelay)
    
    @staticmethod
    @Sim.register_workload_over_hook
    def workload_over_hook(sim: Sim):
        stat_props: SimStats.Props = sim.props.sim_stats
        for sw in stat_props.allStatsWriter:
            sw.inform_workload_is_over()
    
    @staticmethod
    def do_overview_statistics_snapshot(sim: Sim):
        stat_props: SimStats.Props = sim.props.sim_stats
        packet_props: Packet.Props = sim.props.packet
        server_props: Server.Props = sim.props.server
        sff_props: SFF.Props = sim.props.sff
        
        # grab all statistics
        stat_props.overviewStats.add_kv_entry("packet_total_count", packet_props.statsPacketsTotalCount)
        stat_props.overviewStats.add_kv_entry("packet_rejected", packet_props.statsPacketsRejectedSchedule)
        stat_props.overviewStats.add_kv_entry("packet_timeout", packet_props.statsPacketsRejectedProcessingDelay)
        stat_props.overviewStats.add_kv_entry("packet_successful", packet_props.statsPacketsSuccessfulProcessed)
        if packet_props.statsPacketsSuccessfulProcessed > 0:
            tmp = (packet_props.statsPacketsSuccessfulProcessed -
                   packet_props.counter_packet_after_workload_end_in_system_no_timeout)
            if tmp > 0:
                stat_props.overviewStats.add_kv_entry("packet_delivered_delay_avg",
                                                      packet_props.statsRatiosQos /
                                                      tmp)
            else:
                stat_props.overviewStats.add_kv_entry("packet_delivered_delay_avg", 1)
        else:
            stat_props.overviewStats.add_kv_entry("packet_delivered_delay_avg", 1)
        stat_props.overviewStats.add_kv_entry("packet_total_delay", packet_props.statsSumDelay)
        
        idle_time = 0
        for s in server_props.all_servers:
            idle_time += s.stats_idle_time
            if s.is_free():
                idle_time += sim.currentTime - s.stats_last_time_idle
        
        stat_props.overviewStats.add_kv_entry("server_idle_time_total", idle_time)
        stat_props.overviewStats.add_kv_entry("server_idle_time_per_server",
                                              idle_time / len(server_props.all_servers))
        stat_props.overviewStats.add_kv_entry("simulation_time", sim.lastRelevantTime)
        stat_props.overviewStats.add_kv_entry("scheduler_attempts",
                                              sum([s.scheduler.scheduling_attempts for s in
                                                   sff_props.allSFFs.values()]))
        stat_props.overviewStats.add_kv_entry("time_of_last_ingress", sim.last_packet_ingress_time)
    
    @staticmethod
    @Sim.register_simulation_done_hook
    def sim_done_hook(sim: Sim):
        stat_props: SimStats.Props = sim.props.sim_stats
        server_props: Server.Props = sim.props.server
        
        if stat_props.exp_stats is not None:
            stat_props.exp_stats.add_kv_entry("runtime", str(time.time() - sim.start_time))
            stat_props.exp_stats.flush()
        
        if stat_props.serverStats is not None:
            for server in server_props.all_servers:
                stat_props.serverStats.add_entry(server.id, server.processing_cap, len(server.SFIs),
                                                 len(server.SFF_ids),
                                                 server.stats_idle_time +
                                                 (sim.currentTime - server.stats_last_time_idle)
                                                 if server.is_free() else 0)
            stat_props.serverStats.flush()
        
        if stat_props.overviewStats is not None:
            SimStats.do_overview_statistics_snapshot(sim)
            stat_props.overviewStats.flush()
        
        if stat_props.debugStats is not None:
            stat_props.debugStats.flush()
        
        if stat_props.packetStats is not None:
            stat_props.packetStats.flush()
        
        if stat_props.packetCdfStats is not None:
            stat_props.packetCdfStats.prepare_flush()
            stat_props.packetCdfStats.flush()
        
        if stat_props.pollStatistics is not None:
            stat_props.pollStatistics.flush()
        
        if stat_props.serverStats is not None:
            stat_props.serverStats.flush()



#!/usr/bin/env python3
# coding=utf-8
import json
import os
from os import path

from setuptools import setup

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    # read file but remove first 4 lines
    long_description = f.read().split("\n", 4)[4]

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "sfctss", "meta.json"), "r") as f:
    meta = json.load(f)

setup(
    name=meta["title"],
    version=meta["version"],
    description='A Service Function Chain (SFC) Traffic Scheduling Simulator',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mblo/sfctss',
    license=meta["license"],
    author=meta["author"],
    package_data={
        'sfctss': ['meta.json'],
    },
    install_requires=[
        "numpy>=1.18.1",
        "sortedcontainers>=2.1.0",
        "terminaltables>=3.1.0"
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    packages=["sfctss", "sfctss.scheduler", "sfctss.model"],
    python_requires='>=3.6, <4'
)



# coding=utf-8

from .core import *


class GreedyShortestDeadlineFirstScheduler(BaseScheduler):
    
    def __init__(self, sim: Sim,
                 incremental: bool = True, oracle=True,
                 admission_control_threshold_low: float = 0.8,
                 admission_control_threshold_high: float = 1.1
                 ):
        super().__init__(sim=sim,
                         incremental=incremental,
                         oracle=oracle,
                         activate_acp=not oracle,
                         admission_control_threshold_low=admission_control_threshold_low,
                         admission_control_threshold_high=admission_control_threshold_high)
        self.admission_control_threshold_low = admission_control_threshold_low
        self.admission_control_threshold_high = admission_control_threshold_high
        
        assert admission_control_threshold_low < admission_control_threshold_high
        
        if not self.oracle:
            self.rate_estimator: Dict[int, RateEstimator] = {}
    
    def is_always_able_to_build_full_path(self):
        return self.oracle and not self.incremental
    
    def apply_scheduling_logic_for_packet(self, packet: Packet):
        sff_props: SFF.Props = self.sim.props.sff
        
        # get the packet from the scheduler's queue
        if self.requires_queues_per_class():
            popped_packet = self.mySFF.packet_queue_per_class[Flow.get_packet_class_of_packet(packet)].pop()
        else:
            popped_packet = self.mySFF.packet_queue.pop()
        
        assert popped_packet == packet
        
        packet.timeQueueScheduling += packet.get_delta_of_time_mark()
        
        self.mark_time_scheduling_starts()
        
        # this packet starts from this SFF, so search from this point
        p_at_sff: SFF = self.mySFF
        
        if self.sim.DEBUG:
            print(". heuristic scheduler @ {0}".format(p_at_sff))
        
        scheduled_path = []
        
        incremental_path = False
        
        # for each remaining sf type of the requested chain of this packet
        while 0 < len(packet.toBeVisited) and not incremental_path:
            next_sf_type = packet.toBeVisited.pop(0)
            # we need to get a SFI for this corresponding sf type
            # try to get this SFI from the SFF where this packet is currently
            # (p_at_sff)
            
            sff_to_check = sff_props.allSFFs if self.oracle else [p_at_sff.id]
            
            sfi_to_check = []
            #  get a list of all possible SFIs
            # calculate for each of these SFIs the cost value
            best_latency = -1
            for sffIDToAsk in sff_to_check:
                sff_to_ask = sff_props.allSFFs[sffIDToAsk]
                delay_of_sff_connection = 0 if p_at_sff == sff_to_ask else SFF.get_multi_hop_latency_for(self.sim,
                                                                                                         p_at_sff.id,
                                                                                                         sff_to_ask.id)
                if best_latency != -1 and best_latency <= delay_of_sff_connection:
                    # we skip this sff if the latency to this sff is bigger than the best option we found so far
                    continue
                
                if next_sf_type in sff_to_ask.SFIsPerType:
                    for sfi in sff_to_ask.SFIsPerType[next_sf_type]:
                        delay = sfi.get_expected_waiting_time() + sfi.get_expected_processing_time()
                        if p_at_sff != sff_to_ask:
                            delay += SFF.get_delay_of_connection(p_at_sff, sff_to_ask)
                        if best_latency == -1:
                            best_latency = delay
                        best_latency = min(best_latency, delay)
                        sfi_to_check.append((delay, sff_to_ask, sfi))
            
            if len(sfi_to_check) == 0:
                if not self.oracle:
                    raise NameError(f'something is going wrong. I don\'t have any SFI which could serve this packet,'
                                    f'but ACP should have handeled this case!?')
                
                packet.realTimeScheduling += self.get_time_delta_of_scheduling()
                try:
                    raise SchedulingFailure("failed to find a path for {0}".format(
                        str(packet.flow.sfTypeChain)))
                except SchedulingFailure as e:
                    raise e
                finally:
                    packet.reject()
            
            # find the sfi with the lowest delay till packet is processed
            best_sfi = sfi_to_check[0]
            for sfi_tuple in sfi_to_check:
                if sfi_tuple[0] <= best_sfi[0]:
                    # update the best sfi, but for equal values we prefer to stay at our sff
                    if sfi_tuple[0] != best_sfi[0] or sfi_tuple[1] == self.mySFF:
                        best_sfi = sfi_tuple
            
            sff_to_ask = best_sfi[1]
            sfi = best_sfi[2]
            
            if self.sim.DEBUG:
                print(". found a SFI at SFF {0}".format(sff_to_ask))
            
            # do we need to go to another SFF, or do we stay at the
            # current SFF?
            if p_at_sff != sff_to_ask:
                # we need to go to a different SFF, but before,
                # we have to go back to the original SFF, so that
                # if the packet is at a SFI, it goes fist back
                # to the SFF of this SFI, and then to the next SFF
                
                # so go to p_at_sff if the previous path element was
                # a SFI
                if (len(scheduled_path) >
                        0 and scheduled_path[-1][0] == SFI.__name__):
                    scheduled_path.append((SFF.__name__, p_at_sff))
                
                path_to_other_sff = SFF.get_multi_hop_path_for(self.sim, p_at_sff.id, sff_to_ask.id)
                if self.sim.DEBUG:
                    print(". path to this guy contains {0} intermediate SFFs".format(len(path_to_other_sff)))
                for next_sff in path_to_other_sff:
                    scheduled_path.append((SFF.__name__, sff_props.allSFFs[next_sff]))
                
                p_at_sff = sff_to_ask
            
            scheduled_path.append((SFI.__name__, sfi))
            
            # if we are in incremental scheduling mode, we set the incremental_path flag,
            # so that we stop scheduling from here on
            if self.incremental:
                # incremental_path, so we found a different SFF,
                # hence we stop scheduling here
                incremental_path = True
                # and then go back to the SFF for scheduling
                scheduled_path.append((SFF.__name__, sff_to_ask))
        
        if not incremental_path:
            # finally, add the egress SFF
            if p_at_sff.id != packet.flow.desiredEgressSSFid:
                # go back to the sff
                scheduled_path.append((SFF.__name__, p_at_sff))
                path_to_dest = SFF.get_multi_hop_path_for(self.sim, p_at_sff.id, packet.flow.desiredEgressSSFid)
                for sff_id in path_to_dest:
                    scheduled_path.append((SFF.__name__, sff_props.allSFFs[sff_id]))
            else:
                scheduled_path.append((SFF.__name__, sff_props.allSFFs[packet.flow.desiredEgressSSFid]))
        
        if self.sim.DEBUG:
            Packet.debug_print_path(scheduled_path)
        
        packet.fullPath += scheduled_path
        self.scheduling_attempts += 1
        
        if packet.id == self.sim.PACKET_ID_TO_DEBUG:
            print("** debug packet visited the scheduler, current status:")
            Packet.debug_packet(packet)
        
        packet.realTimeScheduling += self.get_time_delta_of_scheduling()
        self.mySFF.handle_packet_from_scheduler(packet)


class LoadUnawareRoundRobinScheduler(BaseScheduler):
    @Sim.register_reset_global_fields
    class Props:
        def __init__(self):
            self.sfi_round_robin_marker = dict()
    
    def __init__(self, sim: Sim, incremental: bool = True, oracle=True):
        super().__init__(sim=sim,
                         incremental=incremental,
                         oracle=oracle,
                         activate_acp=False)
    
    def applies_round_robin(self):
        return True
    
    def is_always_able_to_build_full_path(self):
        return not self.incremental
    
    def apply_scheduling_logic_for_packet(self, packet: Packet):
        sfi_props: SFI.Props = self.sim.props.sfi
        sff_props: SFF.Props = self.sim.props.sff
        
        # get the packet from the scheduler's queue
        if self.requires_queues_per_class():
            popped_packet = self.mySFF.packet_queue_per_class[Flow.get_packet_class_of_packet(packet)].pop()
        else:
            popped_packet = self.mySFF.packet_queue.pop()
        
        assert popped_packet == packet
        
        packet.timeQueueScheduling += packet.get_delta_of_time_mark()
        
        self.mark_time_scheduling_starts()
        
        # this packet starts from this SFF, so search from this point
        p_at_sff_id = self.mySFF.id
        
        scheduled_path = []
        
        # for each remaining sf type of the requested chain of this packet
        while 0 < len(packet.toBeVisited):
            next_sf_type = packet.toBeVisited.pop(0)
            
            # get the sfi which has to serve this packet
            if self.static_sfi_rates_per_sf_cum_weights is None:
                self.update_cum_weights_of_sfi_rates(include_own_sff=True)
            
            assert isinstance(next_sf_type, int)
            if (next_sf_type not in self.static_sfi_rates_per_sf_sorted_sfi
                    or next_sf_type not in self.static_sfi_rates_per_sf_cum_weights
                    or len(self.static_sfi_rates_per_sf_sorted_sfi[next_sf_type]) == 0
                    or len(self.static_sfi_rates_per_sf_cum_weights[next_sf_type]) == 0):
                packet.realTimeScheduling += self.get_time_delta_of_scheduling()
                packet.reject()
                if self.sim.DEBUG:
                    print(f"I'm not aware of any SFI of the required type:{next_sf_type}, so I have to reject the packet")
                raise SchedulingFailure(f"failed to find a path for {packet}")
            
            elif (len(self.static_sfi_rates_per_sf_sorted_sfi[next_sf_type]) !=
                  len(self.static_sfi_rates_per_sf_cum_weights[next_sf_type])):
                raise NameError(f'data structures broken')
            
            target_sfi_id = self.sim.random.choices(self.static_sfi_rates_per_sf_sorted_sfi[next_sf_type],
                                                    cum_weights=self.static_sfi_rates_per_sf_cum_weights[next_sf_type],
                                                    k=1)[0]
            
            target_sff_id = sfi_props.all_sfi[target_sfi_id].sffId
            
            # is this a different sff of where we are currently?
            
            if p_at_sff_id != target_sff_id:
                # we need to go to a different SFF, but before,
                # we have to go back to the original SFF, so that
                # if the packet is at a SFI, it goes fist back
                # to the SFF of this SFI, and then to the next SFF
                # (we need to check this, because thi scheduler supports to schedule the whole path at once)
                
                # so go to p_at_sff if the previous path element was
                # a SFI
                if (len(scheduled_path) >
                        0 and scheduled_path[-1][0] == SFI.__name__):
                    scheduled_path.append((SFF.__name__, sff_props.allSFFs[p_at_sff_id]))
                
                path_to_other_sff = SFF.get_multi_hop_path_for(self.sim, p_at_sff_id, target_sff_id)
                for p in path_to_other_sff:
                    scheduled_path.append((SFF.__name__, sff_props.allSFFs[p]))
                
                p_at_sff_id = target_sff_id
            
            scheduled_path.append((SFI.__name__, sfi_props.all_sfi[target_sfi_id]))
            
            if self.incremental:
                scheduled_path.append((SFF.__name__, sff_props.allSFFs[target_sff_id]))
                break
        
        if self.sim.DEBUG:
            Packet.debug_print_path(scheduled_path)
        
        packet.fullPath += scheduled_path
        self.scheduling_attempts += 1
        
        if packet.id == self.sim.PACKET_ID_TO_DEBUG:
            print("** debug packet visited the scheduler, current status:")
            Packet.debug_packet(packet)
        
        packet.realTimeScheduling += self.get_time_delta_of_scheduling()
        self.mySFF.handle_packet_from_scheduler(packet)
